import re
import pydash as __

def scream_case(value):
  return __.strings.snake_case(value).upper()

def capitalize(txt=''):
  return txt and txt[0].upper() + txt[1:]

def remove_capital(word):
  return f'{word[0].lower()}{word[1:]}'

def contains_digits(d):
  _digits = re.compile('\d')
  return bool(_digits.search(d))

def objPathToTypeParam(fieldPath):
  '''
  from: "brightBeer.name"
  to: "brightBeerName"
  '''
  fieldTags = fieldPath.split('.')
  if len(fieldTags) == 1:
    typeParam = fieldTags[0]
  else:
    typeParam = f'{fieldTags[0]}{capitalize(fieldTags[1])}'
  return typeParam



# from .gstorm import mainClass
from . import cli
from . import helpers
from .BaseGraphQLType import BaseGraphQLType
from .GraphQLType import GraphQLType
from .QueryBuilder import QueryBuilder
from .MutationBuilder import MutationBuilder
from._mutation_runners import (
  create
)

# * Package name:
name = 'gstorm'
# * required here for pypi upload exceptions:
__version__ = "0.4.7"



import os
import click
from ValiotWorker.Logging import (log, LogLevel)
from gstorm import helpers

@click.command()
@click.option('--src', default=None, help='Graphql schema file to clean')
@click.option('--dst', default=None, help='Graphql schema file path to store the new file')
def cleanup(src, dst):
  [src_path, dst_path] = [os.path.join(os.getcwd(), path) for path in [src, dst]]
  helpers.silent_removefile(dst_path)
  with open(src_path, 'r') as src_file:
    with open(dst_path, 'w+') as dst_file:
      for line in src_file:
        is_whitespace = not line.strip()
        if is_whitespace:
          dst_file.write(line)
          continue
        line = line.rstrip() # clean
        line = line.split('#')[0].rstrip() # remove comment
        # append if not empty
        if line:
          dst_file.write(line + '\n')
      dst_file.write('\n')
  log(LogLevel.SUCCESS, 'Done')


from __future__ import annotations
import pydash as __ 
from typing import Optional, List, _GenericAlias, get_type_hints, ForwardRef
from datetime import datetime
from gstorm.helpers.gql import setup_gql
from pprint import pprint
import attr
from dataclasses import dataclass, field
from gstorm.QueryBuilder import QueryBuilder
from gstorm.enums import QueryType
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.helpers.date_helpers import get_local_date, iso8601_to_local_date

def nint(value):
  if value is None:
    return None
  elif type(value) == str:
    return int(value)
  elif type(value) == int:
    return value
  else:
    raise Exception(f"Cannot cast {type(value)} to nullable int")

@attr.s
class GraphQLType(BaseGraphQLType):
  '''Base class for Graphql types,
  has one python's dataclass field per graphql type.
  Plus a __type field with the Type's metadata (opt-out, etc)
  '''
  # __metadata__: str = attr.ib(factory=dict) # attr.ib(default={'opt-out': {}}) # ! MAY BE REQUIRED LATER FOR @opt-out, etc
  __sync__: bool = attr.ib(repr=False, default=False)
  __errors__: List[dict] = attr.ib(repr=False, factory=list)
  id: Optional[int] = attr.ib(default=None, converter=nint, metadata={'unique': {}})
  insertedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True})
  updatedAt: datetime = attr.ib(repr=False, factory=lambda: get_local_date(datetime.now()), converter=iso8601_to_local_date, metadata={'readonly': True})
  
  def __attrs_post_init__(self):
    """[summary]
    """
    self._public_keys = self.get_public_fields()
  def __getitem__(self, item):
    if type(item) == str:
      return self.__getattribute__(item)
    elif type(item) == int:
      key = self._public_keys[item]
      return key, self.__getattribute__(key)
  def __setitem__(self, key, value):
    self.__dict__[key] = value
  @classmethod
  def get_field(cls, name):
    return attr.fields_dict(cls).get(name)
  @classmethod
  def get_fields(cls):
    return attr.fields_dict(cls)
  @classmethod
  def get_public_fields(cls):
    return [key for key in cls.get_fields().keys() if not key.startswith('__')]
  @classmethod
  def get_base_fields(cls):
    list_fields = [
      name
      for name, hint in get_type_hints(cls).items()
      if hint != List[GraphQLType] # if not type List
        and not type(hint) == ForwardRef # GraphQL types are forward-referenced
        and not name.startswith('_')
    ]
    return list_fields
  @classmethod
  def get_list_fields(cls):
    list_fields = [
      name
      for name, hint in get_type_hints(cls).items()
      if hint == List[GraphQLType] # if not type List
    ]
    return list_fields
  @classmethod
  def get_object_fields(cls, typename=None):
    obj_fields = {
      name: hint
      for name, hint in get_type_hints(cls).items()
      if (
        type(hint) == ForwardRef
        # and hint.__forward_arg__ == typename if typename else True # filters by object type
      )
    }
    if typename:
      obj_fields = {
        name: hint
        for name, hint in obj_fields.items()
        if hint.__forward_arg__ == typename
      }
    return list(obj_fields.keys())
  @classmethod
  def get_unique_fields(cls):
    return [
      name
      for name, field in attr.fields_dict(cls).items()
      if 'unique' in field.metadata
    ]
  @classmethod
  def get_field_types(cls):
    return {k:v for k, v in get_type_hints(cls).items() if not k.startswith('_')}
  @classmethod
  def get_metadata(cls, field_name):
    field = attr.fields_dict(cls)[field_name]
    return field.metadata
  def get_unique_identifiers(self):
    return {
      k: self[k]
      for k in self.get_unique_fields()
      if self[k]
    }
  def create(self):
    pass
  def load(self):
    raise NotImplementedError('Child class should implement load method')
  @classmethod
  def query(cls):
    return QueryBuilder(_kind=cls)
  @classmethod
  def query_one(cls):
    return QueryBuilder(cls, QueryType.SINGULAR)
  def is_sync(self):
    return self.__sync__
  def has_errors(self):
    return len(self.__errors__) > 0
  def get_errors(self):
    return self.__errors__
  def update(self, data):
    # Transforms object with selected data
    fields = self.get_fields().keys()
    for key, value in data.items():
      if key in fields:
        self[key] = value
  # ! method aliases
  # query
  q = query
  qm = query
  qp = query # query plural
  # query_one
  q1 = query_one
  qs = query_one # query singular



from __future__ import annotations
from typing import get_type_hints, Dict, List, Union, ForwardRef
from dataclasses import dataclass, field
import json
from datetime import datetime
import inflect
from pygqlc import GraphQLClient
import gstorm
from gstorm.enums import QueryType
from gstorm.helpers.str_handling import remove_capital, contains_digits, capitalize
from gstorm.helpers.date_helpers import get_utc_date, get_iso8601_str
from gstorm.helpers.gql import setup_gql
from gstorm.helpers.typing_helpers import is_enum
from gstorm.helpers.gql_schema_helpers import prettify_query
from gstorm.enums import MutationType, MutationMode

'''

# 1. generar mutations
MUT = MutationBuilder(prog)
# 2. Correr mutations
result, errors = gql.mutate(MUT)
# 3. Update de objeto de Planeacion
if errors:
  prog.set_errors(errors)
  prog.set_sync(false)
else:
  prog.set_sync(true)
  prog.id = result.id
if gstorm.validate(prog):
  send_notification('success')
else:
  send_notification('error')

'''

@dataclass
class MutationBuilder():
  data: Union[List[gstorm.GraphQLType], gstorm.GraphQLType] = None
  # _kind: Union[List[gstorm.GraphQLType], gstorm.GraphQLType] = None # ! IMPLICIT FROM DATA
  mode: MutationMode = MutationMode.BATCH
  _mutation: str = ''
  _children: dict = field(default_factory=dict)
  _isChild: bool = False
  _gql: GraphQLClient = field(default_factory=setup_gql)
  def children(self, *args, **kwargs):
    """Appends child objects to mutation, accepts inputs in several formats:
      children(['room', 'inventories']) # ! DEFAULTS TO SAME MUTATION (CREATE, UPDATE)
      children(
        room = MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
        inventories = MutationType.CREATE, # ! MODE, NO CHILDREN IMPLICIT
      )
      children({
        'room': MutationType.UNIQUE, # ! MODE, NO CHILDREN IMPLICIT
        'inventories': {
          '__mode': MutationType.CREATE, # ! MODE, CHILDREN EXPLICIT
          'tank': MutationType.UPSERT
        },
      })
    
    Returns
    -------
    QueryBuilder
      Same instance that called this method for chaining
    """
    if len(args) > 0:
      if type(args[0]) == str:
        self._children = {
          **self._children,
          **{args[0]: MutationType.LOCAL}
        }
      elif type(args[0]) == list:
        new_children = {ch: MutationType.LOCAL for ch in args[0]}
        self._children = {
          **self._children,
          **new_children
        }
      elif type(args[0]) == dict:
        self._children = {
          **self._children,
          **args[0]
        }
    elif kwargs:
      new_children = {ch: _type for ch, _type in kwargs.items()}
      self._children = {**self._children, **kwargs}
    for key in self._children.keys():
      self._children[key]._isChild = True
    return self
  def pretty_compile(self):
    compiled = self.compile()
    if type(compiled) == list: # sequence of mutations
      return [prettify_query(compiled_item) for compiled_item in compiled]
    else:
      return prettify_query(self.compile())
  def compile(self):
    if type(self.data) == list:
      return self.compile_plural(self.data, self._children)
    else:
      return self.compile_singular(self.data, self._children)
  def compile_plural(self, data, children):
    batch_items = [
      f'create{type(datum).__name__}_{i}: {self.compile_singular(datum, children, batch=True)}'
      for i, datum in enumerate(data)
    ]
    engine = inflect.engine()
    plural_name = engine.plural(remove_capital(type(data[0]).__name__))
    query_name = 'create' + capitalize(plural_name)
    mutation = f'mutation {query_name} {{ ' + " ".join(batch_items) + ' }'
    return mutation
  def compile_singular(self, data, children, batch=False):
    name = f'create{capitalize(type(data).__name__)}'
    if batch:
      _mutation = '$M_NAME $M_PARAMS{ $M_DEF_FIELDS }'
    else:
      _mutation = 'mutation { $M_NAME $M_PARAMS{ $M_DEF_FIELDS } }'
    _m_def_fields = 'successful messages{ field message } result{ $M_FIELDS }'
    fields = [
      k
      for k, hint in get_type_hints(type(data)).items()
      # filter params (only scalar ones)
      if (
        not k.startswith('__')
        and not hint == List[gstorm.GraphQLType]
        and not type(hint) == ForwardRef
      )
    ]
    singular_params = self.compile_singular_params(data)
    child_params = self.compile_child_params(data, children)
    params = singular_params if not child_params else f'{child_params} {singular_params}'
    # compiled result:
    _mutation = _mutation.replace('$M_DEF_FIELDS', _m_def_fields)
    _mutation = _mutation.replace('$M_NAME', name)
    _mutation = _mutation.replace('$M_PARAMS', f'( {params} )')
    _mutation = _mutation.replace('$M_FIELDS', " ".join(fields))
    mutation_singular = _mutation
    # Process list children to include them in the mutation sequence:
    child_plural_mutations = []
    for children_field, _type in children.items():
      if _type == MutationType.CREATE:
        # ! Inject parent into children to include in mutations:
        for child_obj in data[children_field]:
          # ! Example: Tank.get_object_fields('Room')[0]
          parent_field = type(child_obj).get_object_fields(type(data).__name__)[0]
          child_obj[parent_field] = data
        child_plural_mutations.append(self.compile_plural(data[children_field], {parent_field: MutationType.LOCAL}))
    if child_plural_mutations:
      blabla = [mutation_singular, *child_plural_mutations]
      return [mutation_singular, *child_plural_mutations]
    else:
      # ! single mutation, simplest case
      return mutation_singular
  def compile_singular_params(self, data):
    data_type = type(data)
    hints = get_type_hints(data_type)
    params = [
      (k, data[k])
      for k, hint in hints.items()
      # filter params (only scalar ones)
      if (
        not k.startswith('__')
        and not k in ['id', 'insertedAt', 'updatedAt']
        and not hint == List[gstorm.GraphQLType]
        and not type(hint) == ForwardRef
      )
    ]
    return ' '.join([self.gql_param_dumps(k, v) for k, v in params])
  def compile_child_params(self, data, children):
    params = []
    local_children = [
      child
      for child, mtype in children.items()
      if mtype == MutationType.LOCAL
    ]
    for child in local_children:
      child_obj = data[child]
      # get first non-empty unique identifier:
      _id = next(iter(child_obj.get_unique_identifiers()))
      key = f'{child}{capitalize(_id)}'
      params.append(self.gql_param_dumps(key, child_obj[_id]))
    return ' '.join(params)
  def gql_param_dumps(self, k, v):
    """Converts GraphQL parameter pair (key, value) to it's serialized form
    Examples:
    - {'capacity': 10} -> 'capacity: 10'
    - {'enabled': True} -> 'enabled: true'
    - {'name': 'L101'} -> 'name: "L101"'
    - {'status': 'FINISHED'} -> 'status: FINISHED'

    Parameters
    ----------
    k : param key
        left-hand side id for graphql parameter pair
    v : param value
        right-hand side value for graphql parameter pair
    """
    if type(v) == int:
      return f'{k}: {v}'
    if type(v) == float:
      return f'{k}: {v}'
    if type(v) == str:
        # regular case, this means it's a string, include double quotes
        return f'{k}: "{v}"'
    if type(v) == bool:
      return f'{k}: {["false","true"][v]}' # short-hand for inline if
    if type(v) == datetime:
      utc_date = get_utc_date(v)
      return f'{k}: "{get_iso8601_str(utc_date)}"'
    if type(v) == dict:
      quote = '"'
      scape = '\\'
      return f'{k}: "{json.dumps(v).replace(quote, scape + quote)}"'
    if is_enum(v):
      return f'{k}: {v.value}'
  def run(self):
    _mutation = self.compile()
    if type(_mutation) == str:
      # ! SINGLE MUTATION
      data, errors = self._gql.mutate(_mutation)
      if type(self.data) == list:
        # ! BATCH MUTATION
        full_response = {'result': [], 'messages': [], 'successful': []}
        for i, item in enumerate(self.data):
          label = f'create{type(item).__name__}_{i}'
          # data for this specific item:
          item_response = data.get(label)
          # append to full response:
          full_response['successful'].append(item_response.get('successful', False))
          full_response['messages'].extend(item_response.get('messages', []))
          full_response['result'].append(item_response.get('result', {}))
          item.__errors__ = item_response['messages'] if len(item_response.get('messages', [])) > 0 else []
          if 'result' in item_response:
            item.update(item_response['result'])
            item.__sync__ = True
          else:
            item.__errors__.extend([{'message': 'No data for this item'}])
        return full_response
      else:
        # ! SIMPLE MUTATION
        self.data.__errors__ = errors
        if not errors:
          self.data.update(data['result']) # rebuilds object adding data from DB
          self.data.__sync__ = True
        return data
    elif type(_mutation) == list:
      _mutations = _mutation
      # ! SEQUENTIAL MUTATION
      sequence = [self.data]
      sequence.extend([
        self.data[child]
        for child, _type in self._children.items()
        if _type == MutationType.CREATE
      ])
      full_response = {'result': [], 'messages': [], 'successful': []}
      for mutation, seq_obj in zip(_mutations, sequence):
        # ! Run every mutation sequentially
        data, errors = self._gql.mutate(mutation)
        if type(seq_obj) == list:
          # ! SEQUENCE - BATCH MUTATION ITEM
          for i, item in enumerate(seq_obj):
            label = f'create{type(item).__name__}_{i}'
            # data for this specific item:
            item_response = data.get(label)
            # append to full response:
            full_response['successful'].append(item_response.get('successful', False))
            full_response['messages'].extend(item_response.get('messages', []))
            full_response['result'].append(item_response.get('result', {}))
            item.__errors__ = item_response['messages'] if len(item_response.get('messages', [])) > 0 else []
            if 'result' in item_response:
              item.update(item_response['result'])
              item.__sync__ = True
            else:
              item.__errors__.extend([{'message': 'No data for this item'}])
        else:
          # ! SEQUENCE - SINGULAR MUTATION ITEM
          seq_obj.__errors__ = errors
          if not errors:
            seq_obj.update(data['result']) # rebuilds object adding data from DB
            seq_obj.__sync__ = True
      return full_response
    else:
      raise Exception('Compiled mutation should be either a string or a list of strings')


from datetime import datetime
from pytz import timezone
from tzlocal import get_localzone as get_tz

def get_utc_date(date):
  utc = timezone('UTC')
  tz = get_tz() # local timezone
  if not date.tzinfo:
    local_date = tz.localize(date)
  else:
    local_date = date
  utc_date = local_date.astimezone(utc)
  return utc_date

def get_local_date(date):
  tz = get_tz()
  if not date.tzinfo: # if false, already localized, skip
    local_date = tz.localize(date)
  else:
    local_date = date
  return local_date

def iso8601_to_local_date(iso_str):
  if not type(iso_str) == str:
    return iso_str # do nothing if not string, expect it to be a date already
  utc = timezone("UTC")
  utc_date = datetime.strptime(iso_str, "%Y-%m-%dT%H:%M:%SZ")
  utc_date = utc.localize(utc_date)
  tz = get_tz()
  return utc_date.astimezone(tz)

def get_iso8601_str(date):
  return date.strftime("%Y-%m-%dT%H:%M:%SZ")


from typing import Dict, List
from enum import Enum
from dataclasses import dataclass, field
from datetime import datetime
import inflect
from pygqlc import GraphQLClient
from gstorm.BaseGraphQLType import BaseGraphQLType
from gstorm.enums import QueryType
from gstorm.helpers.str_handling import remove_capital, contains_digits
from gstorm.helpers.date_helpers import get_utc_date, get_iso8601_str
from gstorm.helpers.typing_helpers import is_enum
from gstorm.helpers.gql import setup_gql
from gstorm.helpers.gql_schema_helpers import prettify_query

'''
# SIMPLEST
query {
  tanks {
    id
    insertedAt
    updatedAt
  }
}
# PARAMS
query {
  tanks (
    filter: {
      capacity: 10
      name: "hola"
      after: {
        attribute: INSERTED_AT
        date: "2020-01-01T00:00:00Z"
      }
    }
    orderBy: {
      desc: ID
    }
    limit: 10
  ){
    id
    name
    capacity
    insertedAt
    updatedAt
  }
}
'''

@dataclass
class QueryBuilder():
  _kind: BaseGraphQLType = None
  _type: QueryType = QueryType.PLURAL
  _filter: Dict = field(default_factory=dict)
  _findBy: Dict = field(default_factory=dict)
  _orderBy: Dict = field(default_factory=dict)
  _limit: int = 0
  _offset: int = 0
  _page: int = 0
  _query: str = ''
  _children: dict = field(default_factory=dict)
  _isChild: bool = False
  _gql: GraphQLClient = field(default_factory=setup_gql)
  def filter(self, **kwargs):
    self._filter = {**self._filter, **kwargs}
    return self
  def findBy(self, **kwargs):
    self._findBy = kwargs
    return self
  def orderBy(self, **kwargs):
    self._orderBy = kwargs
    return self
  def limit(self, count):
    self._limit = count
    return self
  def offset(self, count):
    self._offset = count
    return self
  def children(self, *args, **kwargs):
    """Appends child objects to query, accepts inputs in three formats:
      child(room=10, inventories=20)
      child({'room':10, 'inventories':20})
      child(**{'room':10, 'inventories':20})
    
    Returns
    -------
    QueryBuilder
      Same instance that called this method for chaining
    """
    if len(args) > 0:
      self._children = {**self._children, **args[0]}
    elif kwargs:
      self._children = {**self._children, **kwargs}
    for key in self._children.keys():
      self._children[key]._isChild = True
    return self
  def paginate(self):
    raise NotImplementedError('Child class should implement paginate method')
  def create(self):
    raise NotImplementedError('Child class should implement create method')
  def update(self):
    raise NotImplementedError('Child class should implement update method')
  def upsert(self):
    raise NotImplementedError('Child class should implement upsert method')
  def delete(self):
    raise NotImplementedError('Child class should implement delete method')
  def nested(self):
    raise NotImplementedError('Child class should implement nested method')
  def apply(self):
    raise NotImplementedError('Child class should implement apply method')
  def get(self):
    """Builds serialized query and returns data from GraphQL Endpoint
    
    Returns
    -------
    GraphQLType
      Instance of the GraphQLType class that created this QueryBuilder
    """
    _query = self.compile()
    data, errors = self._gql.query(_query)
    if self._type == QueryType.SINGULAR:
      if errors:
        instance = self._kind()
        instance.__errors__ = errors
        return instance
      instance = self._kind(**data)
      instance.__sync__ = True
      return instance
    else:
      # QueryType.PLURAL
      if errors:
        instances = [self._kind()]
        instances[0].__errors__ = errors
        return instances
      instances = [self._kind(**datum) for datum in data]
      # mark as synced (AKA: data ok)
      for i, instance in enumerate(instances):
        instances[i].__sync__ = True
      return instances
  def pretty_compile(self):
    return prettify_query(self.compile())
  def compile(self, defaultName=None, isChild=False) -> str:
    """Compiles collected parameters into a serialized query
    
    Parameters
    ----------
    isChild : bool, optional
      If true, looses the external query boilerplate ("query { ... }"), by default False
    
    Returns
    -------
    str
      Serialized query ("query Type { getType{fields...}}")
    """
    self._isChild = isChild
    if self._type == QueryType.PLURAL:
      return self.compile_plural(defaultName=defaultName, isChild=isChild)
    else:
      return self.compile_singular(defaultName=defaultName, isChild=isChild)
  def compile_plural(self, defaultName = None, isChild=False):
    # get query name
    if not defaultName:
      engine = inflect.engine()
      query_name = engine.plural(remove_capital(self._kind.__name__))
    else:
      query_name = defaultName
    _query = 'query { $Q_NAME $Q_PARAMS{ $Q_FIELDS } }' if not self._isChild else '$Q_NAME $Q_PARAMS{ $Q_FIELDS }'
    has_params = any([self._filter, self._orderBy, self._limit])
    _query = _query.replace('$Q_NAME', query_name)
    if has_params:
      _query = _query.replace('$Q_PARAMS', f'( {self.compile_plural_params()} )')
    else:
      _query = _query.replace('$Q_PARAMS', '')
    _query = _query.replace('$Q_FIELDS', self.compile_fields())
    return _query
  def compile_singular(self, defaultName = None, isChild=False):
    # get query name
    if not defaultName:
      query_name = remove_capital(self._kind.__name__)
    else:
      query_name = defaultName
    _query = 'query { $Q_NAME $Q_PARAMS{ $Q_FIELDS } }' if not self._isChild else '$Q_NAME $Q_PARAMS{ $Q_FIELDS }'
    has_params = any([self._findBy])
    _query = _query.replace('$Q_NAME', query_name)
    if has_params:
      _query = _query.replace('$Q_PARAMS', f'( {self.compile_singular_params()} )')
    else:
      _query = _query.replace('$Q_PARAMS', '')
    _query = _query.replace('$Q_FIELDS', self.compile_fields())
    return _query
  def compile_fields(self):
    compiled_child_queries = ''
    if self._children:
      child_queries = [child.compile(defaultName=qname, isChild=True) for qname, child in self._children.items()]
      compiled_child_queries = ' ' + ' '.join(child_queries)
    return " ".join(self._kind.get_base_fields()) + compiled_child_queries
  def compile_plural_params(self):
    # apply filter params
    filter_params = self.gql_param_dumps('filter', self._filter) if self._filter else ''
    # apply orderBy param
    orderby_params = self.gql_param_dumps('orderBy', self._orderBy) if self._orderBy else ''
    # apply limit
    limit_params = f'limit: {self._limit}' if self._limit else ''
    all_params = [filter_params, orderby_params, limit_params]
    used_params = [p for p in all_params if not p == '']
    return f' '.join(used_params)
  def compile_singular_params(self):
    if 'id' in self._findBy:
      k = 'id'
      v = self._findBy['id']
      findby_params = self.gql_param_dumps(k, v)
    else:
      findby_params = self.gql_param_dumps('findBy', self._findBy)
    return findby_params
  def gql_param_dumps(self, k, v):
    """Converts GraphQL parameter pair (key, value) to it's serialized form
    Examples:
    - {'capacity': 10} -> 'capacity: 10'
    - {'enabled': True} -> 'enabled: true'
    - {'name': 'L101'} -> 'name: "L101"'
    - {'status': status.FINISHED} -> 'status: FINISHED'
    - {
        'after': {
          'attribute': 'INSERTED_AT',
          'date': datetime(2020, 4, 4, 9, 12, 47, 1662)
        }
      } -> 'after: { attribute: INSERTED_AT date: "2020-04-04T09:12:47.1662Z" }'
    
    Parameters
    ----------
    k : param key
        left-hand side id for graphql parameter pair
    v : param value
        right-hand side value for graphql parameter pair
    """
    if type(v) == int:
      return f'{k}: {v}'
    if type(v) == float:
      return f'{k}: {v}'
    if type(v) == str:
        # regular case, this means it's a string, include double quotes
        return f'{k}: "{v}"'
    if type(v) == bool:
      return f'{k}: {["false","true"][v]}' # short-hand for inline if
    if type(v) == datetime:
      utc_date = get_utc_date(v)
      return f'{k}: "{get_iso8601_str(utc_date)}"'
    if type(v) == dict:
      subparams = " ".join([self.gql_param_dumps(subk, subv) for subk, subv in v.items()])
      return f'{k}: {{ {subparams} }}'
    if is_enum(v):
      return f'{k}: {v.value}'
  # ! Method aliases:
  # filter
  fil = filter
  # filter sub-aliases:
  # filter(after={'attribute': 'INSERTED_AT', 'date': "2020-02-27T23:01:44Z"})
  def after(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
      _kwargs['date'] = kwargs.get('date')
    else:
      _kwargs = {**kwargs}
    return self.filter(after=_kwargs)
  # filter(before={'attribute': 'INSERTED_AT', 'date': "2020-02-27T23:01:44Z"})
  def before(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
      _kwargs['date'] = kwargs.get('date')
    else:
      _kwargs = {**kwargs}
    return self.filter(before=_kwargs)
  # filter(nullAttribute={ 'isNull': true 'attribute': 'DESCRIPTION' })
  def isNull(self, **kwargs):
    _kwargs = {}
    if 'attr' in kwargs:
      _kwargs['attribute'] = kwargs['attr']
    else:
      _kwargs['attribute'] = kwargs.get('attribute')
    if 'value' in kwargs:
      _kwargs['isNull'] = kwargs['value']
    else:
      _kwargs['isNull'] = kwargs.get('isNull', True) # Default to True
    return self.filter(nullAttribute=_kwargs)
  # findBy
  fb = findBy
  find = findBy
  # orderBy
  ob = orderBy
  order = orderBy
  # limit
  lim = limit
  # offset
  off = offset
  # children
  child = children
  ch = children
  # get
  run = get
  # paginate
  pgt = paginate
  # create
  # update
  # upsert
  # delete
  # nested
  # apply




# ! Referencia para posible pre-procesador de queries

# SOURCE
query = '''
query MyQuery(
  $name: String!
  $capacity: Integer = 1000
){
  gov_tanks: tanks(
    name: $name
    capacity: $capacity
    type: "Gobierno"
  ){
    id
    name
    insertedAt
    location: room {
      id
      name
    }
    inventories (orderBy: { asc: ID } limit: 1) {
      id
      volume
    }
  }
  rooms {
    id
    name
  }
}
'''

# RESULT
parsed_query = {
  '__type': 'Query',
  'label': 'MyQuery',
  'variables': {
    'name': {
      '__type': 'String',
      'required': True,
      'default': None
    },
    'capacity': {
      '__type': 'Integer',
      'required': False,
      'default': 1000
    }
  },
  'operations': [{
    'label': 'gov_tanks',
    'name': 'tanks',
    'params': [
      'name': GraphQLVariable('name')
      'capacity': GraphQLVariable('capacity')
      'type': "Gobierno"
    ],
    'fields': [{
      'label': None,
      'name': 'id',
      'params': [],
      'fields': []
    }, {
      'label': None,
      'name': 'name',
      'params': [],
      'fields': []
    }, {
      'label': None,
      'name': 'insertedAt',
      'params': [],
      'fields': []
    }, {
      'label': 'location',
      'name': 'room',
      'params': [],
      'fields': [{
        'label': None,
        'name': 'id',
        'params': [],
        'fields': []
      }, {
        'label': None,
        'name': 'name',
        'params': [],
        'fields': []
      }]
    }, {
      'label': None,
      'name': 'inventories',
      'params': [
        'orderBy': { 'asc': 'ID' },
        'limit': 1
      ],
      'fields': [{
        'label': None,
        'name': 'id',
        'params': [],
        'fields': []
      }, {
        'label': None,
        'name': 'volume',
        'params': [],
        'fields': []
      }]
    }]
  }, {
    'label': None,
    'name': 'rooms',
    'params': [],
    'fields': [{
      'label': None,
      'name': 'id',
      'params': [],
      'fields': []
    }, {
      'label': None,
      'name': 'name',
      'params': [],
      'fields': []
    }]
  }]
}


from __future__ import annotations
from typing import Union, List
from gstorm.enums import MutationMode
import gstorm

def create(
    data: Union[gstorm.GraphQLType, List[gstorm.GraphQLType]],
    mode: MutationMode = MutationMode.BATCH
  ) -> MutationBuilder:
  """Builds instance of MutationBuilder to sync GraphQLType objects created locally with the remote DB.
  
  Parameters
  ----------
  data : Union[gstorm.GraphQLType, List[gstorm.GraphQLType]]
    GraphQLType instance(s) to create in the DB
  mode : MutationMode, optional
    allows to upload data by different upload mechanisms, by default MutationMode.BATCH
  
  Returns
  -------
  MutationBuilder
    Instance of MutationBuilder class, responsible of building mutation string and communication with DB.
  """
  return gstorm.MutationBuilder(data=data, mode=mode)

def update():
  raise NotImplementedError('gstorm.update not implemented')

def upsert():
  raise NotImplementedError('gstorm.upsert not implemented')

def delete():
  raise NotImplementedError('gstorm.delete not implemented')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
from gstorm import __version__

requirements = [
    # TODO: put package requirements here
    'pygqlc',
    'inflect',
    'pydash',
    'click',
    'pytz',
    'tzlocal',
    'valiotworker',
    'attrs'
]

setup_requirements = [
    # TODO: put setup requirements (distutils extensions, etc.) here
    'twine'
]

test_requirements = [
    # TODO: put package test requirements here
    'pytest',
    'pytest-cov',
    'black'
]

desc = "GraphQL ORM for python (based on pygqlc)"
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='gstorm',
    version=__version__,
    description=desc,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Baruc Almaguer",
    author_email='baruc@valiot.io',
    url='https://github.com/valiot/python-gstorm',
    packages=find_packages(include=['gstorm', 'gstorm.cli']),
    entry_points={
        'console_scripts': [
            'gstorm-cli=gstorm.cli.__main__:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords=['gstorm', 'orm', 'graphql', 'gql'],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



from importlib import import_module
from enum import Enum
from gstorm import BaseGraphQLType

def is_enum(x):
  return Enum in type(x).__mro__

def is_gql_subtype(x):
  return BaseGraphQLType in type(x).__mro__

def enum_cast(x):
  return x.value if is_enum(x) else x


def convert_to(models_path, classname):
  """Returns a casting function to determined type
  Parameters
  ----------
  models_path : str
    String denoting the path to the models directory: "path.to.models"
  classname : str
    Model name (Ej: "Author", "User", etc)
  Returns
  -------
  function
    function to cast to the specified type
  Raises
  ------
  Exception
    When the class was not found in the available models
  """
  models = import_module(models_path)
  def concrete_cast(value):
    if hasattr(models, classname):
      cls = getattr(models, classname)
    else:
      raise Exception(f'"{classname}" model not found inside models module ({models_path})')
    '''casts to specified type'''
    if value is None:
      return None
    if type(value) == dict:
      return cls(**value)
    elif type(value) == str or is_enum(value):
      return cls(value)
    elif is_gql_subtype(value):
      return value
    else:
      return value
  return concrete_cast

def list_convert_to(models_path: str, classname: str):
  """Returns a casting function to determined type
  
  Parameters
  ----------
  models_path : str
    module path where to look for GraphQL models (ej: "tests.models")
  classname : str
    concrete type to convert-to
  
  Returns
  -------
  function
      function to cast to the specified type
  """
  models = import_module(models_path)
  def concrete_cast(values):
    if hasattr(models, classname):
      cls = getattr(models, classname)
    else:
      raise Exception(f'"{classname}" model not found inside models module ({models_path})')
    '''casts list items to specified type'''
    casted_values = []
    for value in values:
      if type(value) == dict:
        casted_values.append(cls(**value))
      elif type(value) == cls:
        init_values = {k: v for k, v in value.__dict__.items() if not k.startswith('_')}
        casted_values.append(cls(**init_values))
    return casted_values
  return concrete_cast

def gql_repr(value):
  if value is None:
    return 'None'
  return f'{type(value).__name__}(...)'

def gql_list_repr(values):
  if values is None:
    return 'None'
  if len(values) == 0:
    return '[]'
  else:
    return f'[{type(values[0]).__name__}({len(values)})]'



import attr

@attr.s
class GraphQLEnum():
  """Wraps an enumeration string to identify it as an enum
  May be needed to identify string literals as enums in GraphQLBuilder.
  Example:
  "L101" -> str
  "INSERTED_AT" -> Enum
  "SUCCESS" -> Enum
  "Gobierno" -> str
  etc
  """
  value: str = attr.ib()
